#pragma once
#ifndef BEAMRAY_HPP
#define BEAMRAY_HPP
#include "frame.hpp"
#include "ray.hpp"

using namespace QOSM::Pose;

namespace QOSM::Propagation{
    enum class BeamType{
        INITIAL = 0,
        REFLECTED,
        TRANSMITTED
    };

    using cplx = std::complex<double>;
    using array2c = std::array<cplx, 2>;
    using array2d = std::array<double, 2>;
    using array2i = std::array<int, 2>;

    struct Beam{             
        Frame   frame;
        double   k0, n, kappa, initialPower;
        cplx     Z_;
        cplx     q1_, q2_, theta_, a0, phi0;
        array2c  alpha_, fresnel_;
        array2d  zRange, z0, zr;
        array2i  bouncesLeft;
        BeamType beamType;
        int      trck_id;
        int      id_triangle;

        Beam();

        /**
         * Instanciate a beam
         *
         * @param freq [Hz] operating frequency. 
         * @param ior (double, double) index of refraction and (positive) extinction coefficient
         * @param polarisation (complex, complex) polarisation coefficients for each principal axis
         * @param w0 (double, double) [m] waist size for each principal axis
         * @param z0 (double, double) [m] waist position for each principal axis
         * @param a0 initial amplitude
         * @param phi0 initial phase
         * @param initPw initial power, computed using polarisation coefficient if 0
         * @param zRange (double, double), range alongside the beam propagation axis
         * @param bounces (int, int), number of reflections and transmittions allowed for the beam
         * @param frame Beam frame w.r.t. the feeding source
         * @param tracking_id Id for following paths tacken by the ray and its children
         */
        Beam(const double &freq, 
             const array2d &ior, 
             const array2c &polarisation, 
             const array2d &w0, 
             const array2d &z0={0., 0.}, 
             const double &a0=1., 
             const double &phi0=0., 
             const double &initPw=0., 
             const array2d &zRange={-1e30, 1e30}, 
             const array2i &bounces={0, 0}, 
             const Frame &frame = Frame(), 
             const  int &tracking_id=-1);

        /**
         * Instanciate a beam
         *
         * @param freq [Hz] operating frequency. 
         * @param ior (double, double) index of refraction and (positive) extinction coefficient
         * @param polarisation (complex, complex) polarisation coefficients for each principal axis
         * @param q (complex, complex) complex parameters for each principal axis
         * @param theta complex angle parameter
         * @param a0 initial amplitude
         * @param phi0 initial phase
         * @param initPw initial power, computed using polarisation coefficient if 0
         * @param zRange (double, double), range alongside the beam propagation axis
         * @param bounces (int, int), number of reflections and transmittions allowed for the beam
         * @param frame Beam frame w.r.t. the feeding source
         * @param tracking_id Id for following paths tacken by the ray and its children
         */
        Beam(const double &freq, 
             const array2d &ior, 
             const array2c &polarisation, 
             const array2c &q, 
             const cplx &theta=0., 
             const double &a0=1., 
             const double &phi0=0., 
             const double &initPw=0., 
             const array2d &zRange={-1e30, 1e30}, 
             const array2i &bounces={0, 0}, 
             const Frame &frame = Frame(), 
             const  int &tracking_id=-1);
        
        /**
         * Compute the electric and magnetic fields for a specific position
         * @param pt position vector
         * @param noZ assume Ez=0
         * @return a pair of complex array, first is the electric field, second is the magnetic field
        */
        std::pair<std::array<cplx, 3>, std::array<cplx, 3>> computeFields(
            const Vec3 &pt,
            const bool noZ=false
        ) const;

        /**
         * Get the ray that carries the beam
         * @return ray
         */
        Ray getRay();
    };
    
    /**
     * Initialise a beam using a complex curvature matrix (flatten).
     * 
     * @param Q_ complex curvatude matrix {Q11, Q12, Q21, Q22}.
     */
    void initBeam(
        Beam& beam, 
        const std::array<cplx, 4>& Q_
    );
    
    /**
     * Initialise a beam 
     * using a complex curvature matrix (flatten) and initial phase and amplitude.
     * 
     * @param Q_ complex curvatude matrix {Q11, Q12, Q21, Q22}.
     * @param ampl initial amplitude of the beam (at the instantiation point).
     * @param phase initial phase of the beam (at the instantiation point).
     */
    void initBeam(
        Beam& beam, 
        const std::array<cplx, 4>& Q_, 
        const cplx& ampl, 
        const cplx& phase
    );
    
    /**
     * Compute the scalar gaussian function (called u) at a specific point
     * 
     * @param Q_ (output) complex curvatude matrix {Q11, Q12, Q21, Q22}
     * @param beam beam used for the computation
     * @param r_pt position vector
     */
    std::array<cplx, 3> gaussianScalarFn(
        std::array<cplx, 4>& Q_, 
        const Beam& beam, 
        const Vec3& r_pt
    );

    using beam_ptr  = std::shared_ptr<QOSM::Propagation::Beam>;
}
#endif
