#pragma once
#ifndef FIELD_HPP
#define FIELD_HPP 1
#include "pose.hpp"
#include <complex>
#include <algorithm>

using namespace QOSM::Pose;

namespace QOSM::EM{
    using cplx = std::complex<double>;
    using MatrixX3c = nc::NdArray<std::complex<double>>;

    const double pi = 3.1415926535897932384626433832795028841971693993751058209749445923078164062862089986280348253421170679;

    /// [m/s] Speed of light
    const double c0 = 299792458.;

    /// void permittivity
    const double epsilon0 = 8.854187817e-12;

    /// void impedence
    const double Z0 = 2 * std::sqrt(pi * 1e-7 / 8.854187817e-12);

    class FieldNx3 : public MatrixX3c{
        protected:
            /// Number of points (number of rows)
            nc::uint32 N;

        public:
            /**
             * Instantiate a Field with 1 row and 3 cols
             */
            FieldNx3() : NdArray(1, 3), N(1){
                fill(cplx{ 0. });
            };

            /**
             * Instantiate a Field with _N rows and 3 cols of a specific value
             * @param _N Number of rows (each row is a point of calculation of the field)
             * @param value Initial value of each component for all rows (default: 0)
             */
            FieldNx3(const size_type _N, cplx value = 0.0) : NdArray(_N, 3), N(_N){
                fill(value);
            }
            
            /**
             * Instantiate a Field with a matrix of complexes (size: Nx3)
             * @param data Matrix of values (each row of 3 values is a point of calculation of the field)
             */
            FieldNx3(const MatrixX3c data) : NdArray(data), N(data.numRows()){};

            /**
             * Get the column of index idx
             * @param idx index of the column to retrieve
             * @return a column of Nx1 elements
             */
            inline auto col(nc::uint32 idx){
                return (*this)(rSlice(), idx).reshape(-1, 1);
            }

            /**
             * Get the row of index idx
             * @param idx index of the row to retrieve
             * @return a row of 1x3 elements
             */
            inline auto row(nc::uint32 idx){
                return (*this)(idx, cSlice()).reshape(1, 3);
            }

            /**
             * Insert into the column idx a column of N values
             * @param idx index of the column
             * @param data an Nx1 column vector
             */
            inline void putCol(nc::uint32 idx, const MatrixX3c &data){
                put(rSlice(), idx, data);
            }

            /**
             * Insert into the row idx a row of 3 values
             * @param idx index of the row
             * @param data an 1x3 row vector
             */
            inline void putRow(nc::uint32 idx, const MatrixX3c &data){
                put(idx, cSlice(), data);
            }
            
            /**
             * Performs the dot product with another field \f$ \textbf{E}_2 \f$, defined with 
             * \f$ \textbf{E} \cdot \textbf{E}_2 = E_{x} \overline{E}_{2,x} + E_{x} \overline{E}_{2,y} + E_{z} \overline{E}_{2,z} \f$
             * @param rhs second field
             * @param data an Nx1 vector of the dot product
             */
            inline nc::NdArray<double> dot(const FieldNx3 &rhs) const{
                return nc::real<double>(nc::sum<cplx>(*this * nc::conj(rhs), nc::Axis::COL)).reshape(-1, 1);
            }
            
            /**
             * Performs the dot product with an Nx3 vectors \f$ \textbf{r} \f$, defined with 
             * \f$ \textbf{E} \cdot \textbf{r} = E_x r_x + E_y r_y + E_z r_z \f$
             * @param rhs vectors Nx3
             * @return an Nx1 vector of the dot product
             */
            inline nc::NdArray<cplx> dot(const Pose::VectorNx3 &rhs) const{
                return nc::sum<cplx>(*this * rhs, nc::Axis::COL).reshape(-1, 1);
            }
            
            /**
             * Performs the cross product with another field \f$ \textbf{E}_2 \f$
             * @param rhs second field
             * @return an Nx3 field of the cross product
             */
            FieldNx3 cross(const FieldNx3 &rhs){
                FieldNx3 lhs = *this;
                FieldNx3 res(N);
                auto r = res.rSlice();
                res.put(r, 0, lhs(r, 1) * rhs(r, 2) - lhs(r, 2) * rhs(r, 1));
                res.put(r, 1, lhs(r, 2) * rhs(r, 0) - lhs(r, 0) * rhs(r, 2));
                res.put(r, 2, lhs(r, 0) * rhs(r, 1) - lhs(r, 1) * rhs(r, 0));
                return res;
            }

            FieldNx3 difference(const FieldNx3 &rhs){
                return FieldNx3(*this - rhs);
            }

            /**
             * Conjugate the field as \f$ \textbf{E} = \Re(\textbf{E}) - \Im(\textbf{E}) \f$
             * @return the conjugated field
             */
            inline FieldNx3 conjugate() const{
                FieldNx3 res(N);
                std::transform(
                    this->cbegin(),
                    this->cend(),
                    res.begin(),
                    [](auto& inValue) -> auto{ return std::conj(inValue); });
                return res;
            }


            /**
             * Compute the magnitude of the field, as \f$ |\textbf{E}| = \sqrt{\textbf{E} \cdot \overline{\textbf{E}}} \f$
             * @param normalise True if the magnitude is normalised with the max value
             * @param db True if the result is expressed in decidel (then the field is normalised)
             * @param db_min set the minimal value of the field if it is expressed in decibel
             * @return the magnitude as an Nx1 array of double
             */
            nc::NdArray<double> magnitude(const bool normalise=false, const bool db=false, const double db_min=-50.) const{
                nc::NdArray<double> m = nc::sqrt(this->dot(*this));
                if (normalise || db){
                    m = m / nc::nanmax(m);
                }
                if (db){
                    m = 20. * nc::log10(m);
                    m.putMask(m < db_min, db_min);
                }
                return m;
            }

            /**
             * Compute the intensity of the field, as \f$ I = |\textbf{E}|^2 = \textbf{E} \cdot \overline{\textbf{E}} \f$
             * @param normalise True if the magnitude is normalised with the max value
             * @param db True if the result is expressed in decidel (then the field is normalised)
             * @param db_min set the minimal value of the field if it is expressed in decibel
             * @return the intensity as an Nx1 array of double
             */
            nc::NdArray<double> intensity(const bool normalise=false, const bool db=false, const double db_min=-50.) const{
                nc::NdArray<double> m = this->dot(*this);
                if (normalise){
                    m = m / nc::nanmax(m);
                }
                if (db){
                    m = 10. * nc::log10(m);
                    m.putMask(m < db_min, db_min);
                }
                return m;
            }

            /**
             * Compute the phase of the field
             * @param dim the component to be used (0=x, 1=y, 2=z, -1=all of them)
             * @param deg True if the result is expressed in degrees
             * @param unwrap True to obtain a phase not comprised between -pi and pi
             * @return The phase as an Nx1 array if a single component is requested; otherwise, an Nx3 array.
             */
            nc::NdArray<double> phase(short dim = -1, bool deg=false, bool unwrap=false) const{
                double factor = deg ? 180./nc::constants::pi : 1.;
                const auto &f = *this;
                nc::NdArray<double> ph;
                if (dim < 0){
                    ph = nc::angle(f) * factor;
                } else {
                    ph = nc::angle(f(f.rSlice(), dim)) * factor;
                }
                if (unwrap)
                    ph = nc::unwrap(ph);
                return ph;
            }

            /**
             * Return the magnitude of the relative difference with a reference field \f$ E_{ref} \f$, as \f$ \frac{|E - E_{ref}|}{max|E_{ref}|} \f$
             */
            nc::NdArray<double> relAbsDifference(const FieldNx3 &ref) const{
                return FieldNx3(*this - ref).magnitude() / nc::max(ref.magnitude())[0];
            }

            /**
             * Return the intensity of the relative difference with a reference field \f$ E_{ref} \f$, as \f$ \frac{|E - E_{ref}|^2}{max|E_{ref}|^2} \f$
             */
            nc::NdArray<double> relIntensityDifference(const FieldNx3 &ref) const{
                return FieldNx3(*this - ref).intensity() / nc::max(ref.intensity())[0];
            }
            
            friend FieldNx3 operator+(FieldNx3 lhs, const cplx &rhs){
                return FieldNx3(lhs + rhs);
            }
            
            friend FieldNx3 operator-(FieldNx3 lhs, const cplx &rhs){
                return FieldNx3(lhs - rhs);
            }

            FieldNx3 &operator+=(const cplx &rhs){
                *this  += rhs;
                return *this;
            }

            FieldNx3 &operator-=(const cplx &rhs){
                *this -= rhs;
                return *this;
            }
    };

    class FieldNx1 : public MatrixX3c{
        protected:
            /// Number of points (number of rows)
            nc::uint32 N;

        public:
            /**
             * Instantiate a Field with 1 row and 3 cols
             */
            FieldNx1() : NdArray(1, 1), N(1){
                fill(cplx{ 0. });
            };

            /**
             * Instantiate a Field with _N rows and 3 cols of a specific value
             * @param _N Number of rows (each row is a point of calculation of the field)
             * @param value Initial value of each component for all rows (default: 0)
             */
            FieldNx1(const size_type _N, cplx value = 0.0) : NdArray(_N, 1), N(_N){
                fill(value);
            }
            
            /**
             * Instantiate a Field with a matrix of complexes (size: Nx1)
             * @param data Vector of N values
             */
            FieldNx1(const MatrixX3c data) : NdArray(data), N(data.numRows()){};
            
            /**
             * Instantiate a Field with a FieldNx3 (size: Nx3)
             * @param data Vector of N values
             * @param dim Set the dimension to use to fill the field
             */
            FieldNx1(const MatrixX3c &data, const nc::uint32 &dim) : NdArray(data.at(data.rSlice(), dim)), N(data.numRows()){};

            FieldNx3 difference(const FieldNx1 &rhs){
                return FieldNx1(*this - rhs);
            }
            
            /**
             * Performs the dot product with another field \f$ \textbf{E}_2 \f$, defined with 
             * \f$ \textbf{E} \cdot \textbf{E}_2 = E_{x} \overline{E}_{2,x} + E_{x} \overline{E}_{2,y} + E_{z} \overline{E}_{2,z} \f$
             * @param rhs second field
             * @param data an Nx1 vector of the dot product
             */
            inline nc::NdArray<double> dot(const FieldNx1 &rhs) const{
                return nc::real<double>(*this * nc::conj(rhs)).reshape(-1, 1);
            }

            /**
             * Conjugate the field as \f$ \textbf{E} = \Re(\textbf{E}) - \Im(\textbf{E}) \f$
             * @return the conjugated field
             */
            inline FieldNx1 conjugate() const{
                FieldNx1 res(N);
                std::transform(
                    this->cbegin(),
                    this->cend(),
                    res.begin(),
                    [](auto& inValue) -> auto{ return std::conj(inValue); });
                return res;
            }
            
            /**
             * Compute the magnitude of the field, as \f$ |\textbf{E}| = \sqrt{\textbf{E} \cdot \overline{\textbf{E}}} \f$
             * @param normalise True if the magnitude is normalised with the max value
             * @param db True if the result is expressed in decidel (then the field is normalised)
             * @param db_min set the minimal value of the field if it is expressed in decibel
             * @return the magnitude as an Nx1 array of double
             */
            nc::NdArray<double> magnitude(const bool normalise=false, const bool db=false, const double db_min=-50.) const{
                nc::NdArray<double> m = nc::sqrt(nc::abs((*this)*(*this)));
                if (normalise || db){
                    m = m / nc::nanmax(m);
                }
                if (db){
                    m = 20. * nc::log10(m);
                    m.putMask(m < db_min, db_min);
                }
                return m;
            }

            /**
             * Compute the intensity of the field, as \f$ I = |\textbf{E}|^2 = \textbf{E} \cdot \overline{\textbf{E}} \f$
             * @param normalise True if the magnitude is normalised with the max value
             * @param db True if the result is expressed in decidel (then the field is normalised)
             * @param db_min set the minimal value of the field if it is expressed in decibel
             * @return the intensity as an Nx1 array of double
             */
            nc::NdArray<double> intensity(const bool normalise=false, const bool db=false, const double db_min=-50.) const{
                nc::NdArray<double> m = nc::abs((*this)*(*this));
                if (normalise){
                    m = m / nc::nanmax(m);
                }
                if (db){
                    m = 10. * nc::log10(m);
                    m.putMask(m < db_min, db_min);
                }
                return m;
            }

            /**
             * Compute the phase of the field
             * @param deg True if the result is expressed in degrees
             * @param unwrap True to obtain a phase not comprised between -pi and pi
             * @return The phase as an Nx1 array if a single component is requested; otherwise, an Nx3 array.
             */
            nc::NdArray<double> phase(bool deg=false, bool unwrap=false) const{
                double factor = deg ? 180./nc::constants::pi : 1.;
                const auto &f = *this;
                nc::NdArray<double> ph;
                ph = nc::angle(f) * factor;
                if (unwrap)
                    ph = nc::unwrap(ph);
                return ph;
            }

            /**
             * Get the row of index idx
             * @param idx index of the row to retrieve
             * @return a row of 1x3 elements
             */
            inline auto row(nc::uint32 idx){
                return (*this)(idx, cSlice()).reshape(1, 3);
            }

            /**
             * Insert into the row idx a row of 3 values
             * @param idx index of the row
             * @param data an 1x3 row vector
             */
            inline void put(nc::uint32 idx, const cplx &data){
                (*this)[idx] = data;
            }

            /**
             * Return the magnitude of the relative difference with a reference field \f$ E_{ref} \f$, as \f$ \frac{|E - E_{ref}|}{max|E_{ref}|} \f$
             */
            nc::NdArray<double> relAbsDifference(const FieldNx1 &ref) const{
                return FieldNx1(*this - ref).magnitude() / nc::max(ref.magnitude())[0];
            }

            /**
             * Return the intensity of the relative difference with a reference field \f$ E_{ref} \f$, as \f$ \frac{|E - E_{ref}|^2}{max|E_{ref}|^2} \f$
             */
            nc::NdArray<double> relIntensityDifference(const FieldNx1 &ref) const{
                return FieldNx1(*this - ref).intensity() / nc::max(ref.intensity())[0];
            }
            
            friend FieldNx1 operator+(FieldNx1 lhs, const cplx &rhs){
                return FieldNx1(lhs + rhs);
            }
            
            friend FieldNx1 operator-(FieldNx1 lhs, const cplx &rhs){
                return FieldNx3(lhs - rhs);
            }

            FieldNx1 &operator+=(const cplx &rhs){
                *this  += rhs;
                return *this;
            }

            FieldNx1 &operator-=(const cplx &rhs){
                *this -= rhs;
                return *this;
            }
    };
}
#endif