#pragma once
#ifndef HORN_HPP
#define HORN_HPP 1
#include <source.hpp>
#include <aperture.hpp>

using namespace QOSM::Propagation;
using namespace QOSM::EM;

namespace QOSM::EM::Source{
    
    /**
     * Definition of a Horn.
     * Implementation of an Electromagnetic Source
     */
    struct Horn : public Source{
        /// @brief  Aperture of the horn
        std::shared_ptr<Aperture> aperture;

        /// @brief  Propagating mode at the aperture
        std::vector<cplx> modes;

        /// @brief  Rotation of the horn along its propagation axis (z)
        double rotZ;

        /**
         * Instanciate a Horn.
         *
         * @param freq_GHz [GHz] Operating frequency.
         */
        Horn(double freq_GHz=0.);

        /**
         * Initialise the Horn with a rectangular aperture.
         *
         * @param modes Modes coefficients that are propagative at the horn aperture.
         * @param Lx [m] width of the aperture.
         * @param Ly [m] height of the aperture.
         * @param rotZdeg [deg] rotation of the horn aperture w.r.t. the optical axis (local z).
         * @param numPtsAperturePerAxis Number of point per axis to sample the horn aperture surface.
         * @return the field at each position.
         */
        void init(
            std::vector<cplx> &modes, 
            const double &Lx, 
            const double &Ly, 
            const double &rotZdeg, 
            const size_t numPtsAperturePerAxis
        );
        
        /**
         * Compute electrical field radiated by the horn aperture.
         *
         * @param r_pts_src Position vector where to compute the e-field.
         * @return the field at each position.
         */
        FieldNx3 eField(const VectorNx3 &r_pts_src) const;

        /**
         * Compute magnetic field radiated by the horn aperture.
         *
         * @param r_pts_src Position vector where to compute the h-field.
         * @return the field at each position.
         */
        FieldNx3 hField(const VectorNx3 &r_pts_src) const;

        /**
         * Compute electric, magnetic fields and Poyinting vectors.
         *
         * @param r_pts_src Position vector where to compute the fields.
         * @return a tuple of electrical and magnetic fields and Poyinting vectors at each position.
         */
        std::tuple<FieldNx3, FieldNx3, VectorNx3> fields(const VectorNx3 &r_pts_src) const;

        /**
         * Compute electric, magnetic fields and Poyinting vectors at the horn aperture.
         *
         * @return a tuple of electrical and magnetic fields and Poyinting vectors at the horn aperture.
         */
        std::tuple<FieldNx3, FieldNx3, VectorNx3> apertureFields() const;       
    };
}
#endif