#pragma once
#ifndef SOURCE_HPP
#define SOURCE_HPP
#include <vector>
#include <beam.hpp>
#include <field.hpp>

using namespace QOSM::Propagation;
using namespace QOSM::EM;

namespace QOSM::EM::Source{
    /**
     * Interface for a Electromagnetic Source
     */
    struct Source{

        /// [Hz] Operating frequency
        double freq;
        /// [1/m] Free-Space Wavenumber
        double k0;
        /// [m] Free-Space wavelengh
        double lambda0;

        virtual ~Source() = default;

        /**
         * Create a Source
         * @param freq_GHz [GHz] Operating frequency
         */
        Source(double freq_GHz);

        /**
         * Compute radiated electric field
         * @param r_pts_src [Nx3] Position vectors of the points where the field is to be calculated, written in the source frame
         * @return [Nx3] Electric field
         */
        virtual FieldNx3 eField(const VectorNx3 &r_pts_src) const;

        /**
         * Compute radiated magnetic field
         * @param r_pts_src [Nx3] Position vectors of the points where the field is to be calculated, written in the source frame
         * @return [Nx3] Magnetic field
         */
        virtual FieldNx3 hField(const VectorNx3 &r_pts_src) const;

        /**
         * Compute radiated fields and Poyinting Vector
         * @param r_pts_src [Nx3] Position vectors of the points where the field is to be calculated, written in the source frame
         * @return A tuple of ([Nx3] Electric field, [Nx3] Magnetic field, [Nx3] Poyinting Vectors)
         */
        virtual std::tuple<FieldNx3, FieldNx3, VectorNx3> fields(const VectorNx3 &r_pts_src) const;

        /**
         * Compute radiated fields and Poyinting Vector, for one specific direction
         * @param r_pts_src [Nx3] Position vectors of the points where the field is to be calculated, written in the source frame
         * @param z_probe_src [1x3] Probe direction written in source frame
         * @return A tuple of ([Nx3] Electric field, [Nx3] Magnetic field, [Nx3] Poyinting Vectors)
         */
        virtual std::tuple<FieldNx3, FieldNx3, VectorNx3> orientedFields(const VectorNx3 &r_pts_src, const Vec3 &z_probe_src) const;
    };
}
#endif