#pragma once
#ifndef VSRC_HPP
#define VSRC_HPP 1
#include <vector>
#include <source.hpp>
#include <field.hpp>

using namespace QOSM::Propagation;
using namespace QOSM::EM;

namespace QOSM::EM::Source{

    /**
     * Implementation of a Source using vectorial gaussian beams
     *
     */
    struct VirtualSource : public Source{
        /// List of gaussian beams
        std::vector<Beam> beams;

        /// Set the assumption about the z-component of the electrical field (True if Ez=0) 
        bool assumeEzIs0;

        /**
         * Instanciate a VirtualSource.
         *
         * @param freq_GHz [GHz] Operating frequency.
         */
        VirtualSource(double freq_GHz=0.);

        /**
         * Initialise the source using a list of beams.
         * @param list_beams List of beams.
         */
        void init(std::vector<Beam> &list_beams);

        /** 
         * Extend the VirtualSource with the beams from another VSRC
         * @param vsrc The second virtual source to be used
         */
        VirtualSource extends(const VirtualSource &vsrc);

        /**
         * Set the assumtion "Ez = 0" for each gaussian beam
        */
        void setEzIs0Assumption(bool doAssume){
            assumeEzIs0 = doAssume;
        }

        /**
         * Compute electrical field at each position.
         *
         * @param r_pts_src Position vector where to compute the e-field.
         * @return the field at each position
         */
        FieldNx3 eField(const VectorNx3 &r_pts_src) const;

        /**
         * Compute magnetic field at each position.
         *
         * @param r_pts_src Position vector where to compute the e-field.
         * @return the field at each position
         */
        FieldNx3 hField(const VectorNx3 &r_pts_src) const;

        /**
         * Compute electric, magnetic fields and Poyinting vectors.
         *
         * @param r_pts_src Position vector where to compute the fields.
         * @return a tuple of electrical and magnetic fields and Poyinting vectors at each position.
         */
        std::tuple<FieldNx3, FieldNx3, VectorNx3> fields(const VectorNx3 &r_pts_src) const;
        
        std::tuple<FieldNx3, FieldNx3, VectorNx3> orientedFields(const VectorNx3 &r_pts_src, const Vec3 &z_probe_src) const;

        VirtualSource filter(const BeamType &type);

        /**
         * Split the beams between initial beam and beam generated after beam tracing.
         * @return a pair of VirtualSource, first is for generated beams, second for initial beams.
         */
        std::pair<VirtualSource, VirtualSource> split();
    };
}
#endif